from pydbus import SessionBus, connect
import os

DBUS_SESSION_BUS_ADDRESS = os.getenv("DBUS_SESSION_BUS_ADDRESS")

with connect(DBUS_SESSION_BUS_ADDRESS) as bus:
	bus.dbus

del bus._dbus
try:
	bus.dbus
	assert(False)
except RuntimeError:
	pass

with SessionBus() as bus:
	pass

# SessionBus() and SystemBus() are not closed automatically, so this should work:
bus.dbus

with bus.request_name("net.lew21.Test"):
	pass

with bus.request_name("net.lew21.Test"):
	pass

with bus.request_name("net.lew21.Test"):
	try:
		bus.request_name("net.lew21.Test")
		assert(False)
	except RuntimeError:
		pass

with bus.watch_name("net.lew21.Test"):
	pass

with bus.subscribe(sender="net.lew21.Test"):
	pass



#!/usr/bin/env python

from gi.repository import GLib
from pydbus import SessionBus
from pydbus.generic import signal

class Notifications(object):
	"""
	<node>
		<interface name="org.freedesktop.Notifications">
			<signal name="NotificationClosed">
				<arg direction="out" type="u" name="id"/>
				<arg direction="out" type="u" name="reason"/>
			</signal>
			<signal name="ActionInvoked">
				<arg direction="out" type="u" name="id"/>
				<arg direction="out" type="s" name="action_key"/>
			</signal>
			<method name="Notify">
				<arg direction="out" type="u"/>
				<arg direction="in" type="s" name="app_name"/>
				<arg direction="in" type="u" name="replaces_id"/>
				<arg direction="in" type="s" name="app_icon"/>
				<arg direction="in" type="s" name="summary"/>
				<arg direction="in" type="s" name="body"/>
				<arg direction="in" type="as" name="actions"/>
				<arg direction="in" type="a{sv}" name="hints"/>
				<arg direction="in" type="i" name="timeout"/>
			</method>
			<method name="CloseNotification">
				<arg direction="in" type="u" name="id"/>
			</method>
			<method name="GetCapabilities">
				<arg direction="out" type="as" name="caps"/>
			</method>
			<method name="GetServerInformation">
				<arg direction="out" type="s" name="name"/>
				<arg direction="out" type="s" name="vendor"/>
				<arg direction="out" type="s" name="version"/>
				<arg direction="out" type="s" name="spec_version"/>
			</method>
		</interface>
	</node>
	"""

	NotificationClosed = signal()
	ActionInvoked = signal()

	def Notify(self, app_name, replaces_id, app_icon, summary, body, actions, hints, timeout):
		print("Notification: {} {} {} {} {} {} {} {}".format(app_name, replaces_id, app_icon, summary, body, actions, hints, timeout))
		return 4 # chosen by fair dice roll. guaranteed to be random.

	def CloseNotification(self, id):
		pass

	def GetCapabilities(self):
		return []

	def GetServerInformation(self):
		return ("pydbus.examples.notifications_server", "pydbus", "?", "1.1")

bus = SessionBus()
bus.publish("org.freedesktop.Notifications", Notifications())
loop = GLib.MainLoop()
loop.run()



from __future__ import print_function
from pydbus import SessionBus
from gi.repository import GLib
import subprocess
from time import sleep

loop = GLib.MainLoop()

subprocess.Popen("gnome-music")
sleep(5)
print("Waiting for GNOME Music to start...")

b = SessionBus()
m = b.get("org.mpris.MediaPlayer2.GnomeMusic", "/org/mpris/MediaPlayer2")

m.PropertiesChanged.connect(print)

m.ActivatePlaylist(m.GetPlaylists(0, 5, "Alphabetical", 0)[0][0])

m.Play()
sleep(1)
assert(m.PlaybackStatus == "Playing")

m.Pause()
assert(m.PlaybackStatus == "Paused")

m.Play()
assert(m.PlaybackStatus == "Playing")

t = m.Metadata["xesam:title"]

m.Next()

assert(m.Metadata["xesam:title"] != t)

m.Previous()

assert(m.Metadata["xesam:title"] == t)

oldLoopStatus = m.LoopStatus

m.LoopStatus = "None"
assert(m.LoopStatus == "None")

m.LoopStatus = "Track"
assert(m.LoopStatus == "Track")

m.LoopStatus = oldLoopStatus

GLib.timeout_add_seconds(2, lambda: loop.quit())
loop.run()



from gi.repository import GLib

class ProxyProperty(object):
	def __init__(self, iface_name, property):
		self._iface_name = iface_name
		self.__name__ = property.attrib["name"]
		self.__qualname__ = self._iface_name + "." + self.__name__

		self._type = property.attrib["type"]
		access = property.attrib["access"]
		self._readable = access.startswith("read")
		self._writeable = access.endswith("write")
		self.__doc__ = "(" + self._type + ") " + access

	def __get__(self, instance, owner):
		if instance is None:
			return self

		if not self._readable:
			raise AttributeError("unreadable attribute")

		return instance._object["org.freedesktop.DBus.Properties"].Get(self._iface_name, self.__name__)

	def __set__(self, instance, value):
		if instance is None or not self._writeable:
			raise AttributeError("can't set attribute")

		instance._object["org.freedesktop.DBus.Properties"].Set(self._iface_name, self.__name__, GLib.Variant(self._type, value))

	def __repr__(self):
		return "<property " + self.__qualname__ + " at 0x" + format(id(self), "x") + ">"



from collections import OrderedDict
from inspect import getargspec

class _empty:
	pass

class Signature:
	empty = _empty

	def __init__(self, parameters=None, return_annotation=_empty):
		self.parameters = OrderedDict(((param.name, param) for param in parameters))
		self.return_annotation = return_annotation

class Parameter:
	empty = _empty

	POSITIONAL_ONLY = 0
	POSITIONAL_OR_KEYWORD = 1
	KEYWORD_ONLY = 999

	def __init__(self, name, kind, default=_empty, annotation=_empty):
		self.name = name
		self.kind = kind
		self.annotation = annotation

def signature(f):
	parameters = [Parameter(arg, Parameter.POSITIONAL_OR_KEYWORD) for arg in getargspec(f).args]
	return Signature(parameters = parameters)



from gi.repository import GLib
from collections import namedtuple

AuthorizationResult = namedtuple("AuthorizationResult", "is_authorized is_challenge details")

class MethodCallContext(object):
	def __init__(self, gdbus_method_invocation):
		self._mi = gdbus_method_invocation

	@property
	def bus(self):
		return self._mi.get_connection().pydbus

	@property
	def sender(self):
		return self._mi.get_sender()

	@property
	def object_path(self):
		return self._mi.get_object_path()

	@property
	def interface_name(self):
		return self._mi.get_interface_name()

	@property
	def method_name(self):
		return self._mi.get_method_name()

	def check_authorization(self, action_id, details, interactive=False):
		return AuthorizationResult(*self.bus.polkit_authority.CheckAuthorization(('system-bus-name', {'name': GLib.Variant("s", self.sender)}), action_id, details, 1 if interactive else 0, ''))

	def is_authorized(self, action_id, details, interactive=False):
		return self.check_authorization(action_id, details, interactive).is_authorized



from pydbus import SessionBus
from gi.repository import GLib
from threading import Thread
import sys

done = 0
loop = GLib.MainLoop()

class TestObject(object):
	'''
<node>
	<interface name='net.lew21.pydbus.tests.Iface1'>
		<method name='Method1'>
			<arg type='s' name='response' direction='out'/>
		</method>
	</interface>
	<interface name='net.lew21.pydbus.tests.Iface2'>
		<method name='Method2'>
			<arg type='s' name='response' direction='out'/>
		</method>
	</interface>
</node>
	'''
	def Method1(self):
		global done
		done += 1
		if done == 2:
			loop.quit()
		return "M1"

	def Method2(self):
		global done
		done += 1
		if done == 2:
			loop.quit()
		return "M2"

bus = SessionBus()

with bus.publish("net.lew21.pydbus.tests.expose_multiface", TestObject()):
	remote = bus.get("net.lew21.pydbus.tests.expose_multiface")

	def t1_func():
		print(remote.Method1())
		print(remote.Method2())

	t1 = Thread(None, t1_func)
	t1.daemon = True

	def handle_timeout():
		print("ERROR: Timeout.")
		sys.exit(1)

	GLib.timeout_add_seconds(2, handle_timeout)

	t1.start()

	loop.run()

	t1.join()



from pydbus import SystemBus
from gi.repository import GLib
import logging

loop = GLib.MainLoop()

class TestObject(object):
	dbus = '''
<node>
	<interface name='net.lew21.pydbus.PolkitExample'>
		<method name='TestAuth'>
			<arg type='b' name='interactive' direction='in'/>
			<arg type='s' name='response' direction='out'/>
		</method>
	</interface>
</node>
	'''
	def TestAuth(self, interactive, dbus_context):
		if dbus_context.is_authorized('org.freedesktop.policykit.exec', {'polkit.icon': 'abcd', 'aaaa': 'zzzz'}, interactive=interactive):
			return "OK"
		else:
			return "Forbidden"

with SystemBus() as bus:
	with bus.publish("net.lew21.pydbus.PolkitExample", TestObject()):
		logging.info("Started.")
		loop.run()



from __future__ import print_function
from pydbus.identifier import filter_identifier
import sys

tests = [
	("abc", "abc"),
	("a_bC", "a_bC"),
	("a-b_c", "a_b_c"),
	("a@bc", "abc"),
	("!@#$%^&*", ""),
]

ret = 0
for input, output in tests:
	if not filter_identifier(input) == output:
		print("ERROR: filter(" + input + ") returned: " + filter_identifier(input), file=sys.stderr)
		ret = 1

sys.exit(ret)



#!/usr/bin/env python

# Based on http://stackoverflow.com/questions/22390064/use-dbus-to-just-send-a-message-in-python

# Python DBUS Test Server
# runs until the Quit() method is called via DBUS

from gi.repository import GLib
from pydbus import SessionBus

loop = GLib.MainLoop()

class MyDBUSService(object):
	"""
		<node>
			<interface name='net.lew21.pydbus.ClientServerExample'>
				<method name='Hello'>
					<arg type='s' name='response' direction='out'/>
				</method>
				<method name='EchoString'>
					<arg type='s' name='a' direction='in'/>
					<arg type='s' name='response' direction='out'/>
				</method>
				<method name='Quit'/>
			</interface>
		</node>
	"""

	def Hello(self):
		"""returns the string 'Hello, World!'"""
		return "Hello, World!"

	def EchoString(self, s):
		"""returns whatever is passed to it"""
		return s

	def Quit(self):
		"""removes this object from the DBUS connection and exits"""
		loop.quit()

bus = SessionBus()
bus.publish("net.lew21.pydbus.ClientServerExample", MyDBUSService())
loop.run()



from pydbus import SessionBus, Variant
from gi.repository import GLib
from threading import Thread
import sys

done = 0
loop = GLib.MainLoop()

class TestObject(object):
	'''
<node>
	<interface name='net.lew21.pydbus.tests.publish_properties'>
		<property name="Foobar" type="s" access="readwrite"/>
		<property name="Foo" type="s" access="read"/>
		<property name="Bar" type="s" access="write"/>
		<method name='Quit'/>
	</interface>
</node>
	'''
	def __init__(self):
		self.Foo = "foo"
		self.Foobar = "foobar"

	def Quit(self):
		loop.quit()

bus = SessionBus()

with bus.publish("net.lew21.pydbus.tests.publish_properties", TestObject()):
	remote = bus.get("net.lew21.pydbus.tests.publish_properties")
	remote_iface = remote['net.lew21.pydbus.tests.publish_properties']

	def t1_func():
		for obj in [remote, remote_iface]:
			assert(obj.Foo == "foo")
			assert(obj.Foobar == "foobar")
			obj.Foobar = "barfoo"
			assert(obj.Foobar == "barfoo")
			obj.Foobar = "foobar"
			assert(obj.Foobar == "foobar")
			obj.Bar = "rab"

		remote.Foobar = "barfoo"

		try:
			remote.Get("net.lew21.pydbus.tests.publish_properties", "Bar")
			assert(False)
		except GLib.GError:
			pass
		try:
			remote.Set("net.lew21.pydbus.tests.publish_properties", "Foo", Variant("s", "haxor"))
			assert(False)
		except GLib.GError:
			pass
		assert(remote.GetAll("net.lew21.pydbus.tests.publish_properties") == {'Foobar': 'barfoo', 'Foo': 'foo'})
		remote.Quit()

	t1 = Thread(None, t1_func)
	t1.daemon = True

	def handle_timeout():
		print("ERROR: Timeout.")
		sys.exit(1)

	GLib.timeout_add_seconds(2, handle_timeout)

	t1.start()

	loop.run()

	t1.join()



from pydbus import SessionBus
from gi.repository import GLib
from threading import Thread
import sys

done = 0
loop = GLib.MainLoop()

class TestObject(object):
	'''
<node>
	<interface name='net.lvht.Foo1'>
		<method name='HelloWorld'>
			<arg type='s' name='a' direction='in'/>
			<arg type='i' name='b' direction='in'/>
			<arg type='s' name='response' direction='out'/>
		</method>
	</interface>
</node>
	'''
	def __init__(self, id):
		self.id = id

	def HelloWorld(self, a, b):
		res = self.id + ": " + a + str(b)
		global done
		done += 1
		if done == 2:
			loop.quit()
		print(res)
		return res

bus = SessionBus()

with bus.publish("net.lew21.pydbus.Test", TestObject("Main"), ("Lol", TestObject("Lol"))):
	remoteMain = bus.get("net.lew21.pydbus.Test")
	remoteLol = bus.get("net.lew21.pydbus.Test", "Lol")

	def t1_func():
		print(remoteMain.HelloWorld("t", 1))

	def t2_func():
		print(remoteLol.HelloWorld("t", 2))

	t1 = Thread(None, t1_func)
	t2 = Thread(None, t2_func)
	t1.daemon = True
	t2.daemon = True

	def handle_timeout():
		print("ERROR: Timeout.")
		sys.exit(1)

	GLib.timeout_add_seconds(2, handle_timeout)

	t1.start()
	t2.start()

	loop.run()

	t1.join()
	t2.join()



#!/usr/bin/env python

# Based on http://stackoverflow.com/questions/22390064/use-dbus-to-just-send-a-message-in-python

# Python script to call the methods of the DBUS Test Server

from pydbus import SessionBus

#get the session bus
bus = SessionBus()
#get the object
the_object = bus.get("net.lew21.pydbus.ClientServerExample")

#call the methods and print the results
reply = the_object.Hello()
print(reply)

reply = the_object.EchoString("test 123")
print(reply)

the_object.Quit()



from pydbus import SystemBus

bus = SystemBus()

systemd = bus.get(".systemd1")
#systemd = bus.get("org.freedesktop.systemd1")

manager = systemd[".Manager"]
#manager = systemd["org.freedesktop.systemd1.Manager"]
#manager = systemd # works but may break if systemd adds another interface

import sys

try:
	if len(sys.argv) < 2:
		for unit in manager.ListUnits():
			print(unit)
	else:
		if sys.argv[1] == "--help":
			help(manager)
		else:
			command = sys.argv[1]
			command = "".join(x.capitalize() for x in command.split("-"))
			result = getattr(manager, command)(*sys.argv[2:])

			for var in result:
				if type(var) == list:
					for line in var:
						print(line)
				else:
					print(var)
except Exception as e:
	print(e)

"""
Examples:

python -m pydbus.examples.systemctl
sudo python -m pydbus.examples.systemctl start-unit cups.service replace
"""



#!/usr/bin/env python3

from setuptools import setup

with open('README.rst') as f:
	readme = f.read()

setup(
	name = "pydbus",
	version = "0.6.0",
	description = "Pythonic DBus library",
	long_description = readme,
	author = "Linus Lewandowski",
	author_email = "linus@lew21.net",
	url = "https://github.com/LEW21/pydbus",
	keywords = "dbus",
	license = "LGPLv2+",

	packages = ["pydbus"],
	package_data = {"": ["LICENSE"]},
	package_dir = {"pydbus": "pydbus"},
	zip_safe = True,
	classifiers = [
		'Development Status :: 5 - Production/Stable',
		'Intended Audience :: Developers',
		'Natural Language :: English',
		'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
		'Programming Language :: Python',
		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.3',
		'Programming Language :: Python :: 3.4',
		'Programming Language :: Python :: 3.5',
		'Programming Language :: Python :: 2',
		'Programming Language :: Python :: 2.7'
	]
)



from gi.repository import Gio
from .proxy import ProxyMixin
from .request_name import RequestNameMixin
from .bus_names import OwnMixin, WatchMixin
from .subscription import SubscriptionMixin
from .registration import RegistrationMixin
from .publication import PublicationMixin

def pydbus_property(self):
	try:
		return self._pydbus
	except AttributeError:
		self._pydbus = Bus(self)
		return self._pydbus

Gio.DBusConnection.pydbus = property(pydbus_property)

def bus_get(type):
	return Gio.bus_get_sync(type, None).pydbus

def connect(address):
	c = Gio.DBusConnection.new_for_address_sync(address, Gio.DBusConnectionFlags.AUTHENTICATION_CLIENT | Gio.DBusConnectionFlags.MESSAGE_BUS_CONNECTION, None, None)
	c.pydbus.autoclose = True
	return c.pydbus

class Bus(ProxyMixin, RequestNameMixin, OwnMixin, WatchMixin, SubscriptionMixin, RegistrationMixin, PublicationMixin):
	Type = Gio.BusType

	def __init__(self, gio_con):
		self.con = gio_con
		self.autoclose = False

	def __enter__(self):
		return self

	def __exit__(self, exc_type, exc_value, traceback):
		if self.autoclose:
			self.con.close_sync(None)

	@property
	def dbus(self):
		try:
			return self._dbus
		except AttributeError:
			self._dbus = self.get(".DBus")[""]
			return self._dbus

	@property
	def polkit_authority(self):
		try:
			return self._polkit_authority
		except AttributeError:
			self._polkit_authority = self.get(".PolicyKit1", "Authority")[""]
			return self._polkit_authority

def SystemBus():
	return bus_get(Bus.Type.SYSTEM)

def SessionBus():
	return bus_get(Bus.Type.SESSION)



"""Generic programming utilities.

Utilities implemented in this file are not dependent
on dbus, they can be used everywhere.
"""

class subscription(object):
	__slots__ = ("callback_list", "callback")

	def __init__(self, callback_list, callback):
		self.callback_list = callback_list
		self.callback = callback
		self.callback_list.append(callback)

	def unsubscribe(self):
		self.callback_list.remove(self.callback)
		self.callback_list = None
		self.callback = None

	def disconnect(self):
		"""An alias for unsubscribe()"""
		self.unsubscribe()

	def __enter__(self):
		return self

	def __exit__(self, exc_type, exc_value, traceback):
		if not self.callback is None:
			self.unsubscribe()

class bound_signal(object):
	__slots__ = ("__signal__", "__self__") # bound method uses ("__func__", "__self__")

	def __init__(self, signal, instance):
		self.__signal__ = signal
		self.__self__ = instance

	@property
	def callbacks(self):
		return self.__signal__.map[self.__self__]

	def connect(self, callback):
		"""Subscribe to the signal."""
		return self.__signal__.connect(self.__self__, callback)

	def emit(self, *args):
		"""Emit the signal."""
		self.__signal__.emit(self.__self__, *args)

	def __call__(self, *args):
		"""Emit the signal."""
		self.emit(*args)

	def __repr__(self):
		return "<bound signal " + self.__signal__.__qualname__ + " of " + repr(self.__self__) + ">"

class signal(object):
	"""Static signal object

	You're expected to set it as a class property::

		class A:
			SomethingHappened = signal()

	Declared this way, it can be used on class instances
	to connect signal observers::

		a = A()
		a.SomethingHappened.connect(func)

	and emit the signal::

		a.SomethingHappened()

	You may pass any parameters to the emiting function
	- they will be forwarded to all subscribed callbacks.
	"""

	def __init__(self):
		self.map = {}
		self.__qualname__ = "<anonymous signal>" # function uses <lambda> ;)
		self.__doc__ = "Signal."

	def connect(self, object, callback):
		"""Subscribe to the signal."""
		return subscription(self.map.setdefault(object, []), callback)

	def emit(self, object, *args):
		"""Emit the signal."""
		for cb in self.map.get(object, []):
			cb(*args)

	def __get__(self, instance, owner):
		if instance is None:
			return self

		return bound_signal(self, instance)

	def __set__(self, instance, value):
		raise AttributeError("can't set attribute")

	def __repr__(self):
		return "<signal " + self.__qualname__ + " at 0x" + format(id(self), "x") + ">"

bound_method = type(signal().emit) # TODO find a prettier way to get this type
